import re
import subprocess
import zipfile
from pathlib import Path, PurePosixPath
from string import Template
from typing import Dict, List, Union

import giturlparse
import requests
from fastcore.script import Param, call_parse, store_true
from loguru import logger

import pymemri

TEMPLATE_URL = (
    "https://gitlab.memri.io/memri/plugin-templates/-/archive/dev/plugin-templates-dev.zip"
)
TEMPLATE_BASE_PATH = "plugin-templates-dev"


# If the owner of the repository is one of these groups, the CLI requires an additional `user` argument
GITLAB_GROUPS = ["memri", "plugins"]


def get_remote_url():
    path = Path(".")
    url = subprocess.getoutput("git config --get remote.origin.url")
    if not url:
        raise ValueError(
            f"You can only run this from a initialized gitlab repository, and '{path}' is not an initialized git repository"
        )
    parsed = giturlparse.parse(url)
    repo_url = parsed.url2https
    if repo_url.endswith(".git"):
        repo_url = repo_url[:-4]
    return repo_url


def infer_git_info(url):
    parsed = giturlparse.parse(url)
    return parsed.owner, parsed.repo


def download_file(url, fname=None):
    cert_path = Path(pymemri.__file__).parent / "cert" / "gitlab.memri.io.pem"
    r = requests.get(url, stream=True, verify=cert_path)
    fname = url.rsplit("/", 1)[1] if fname is None else fname
    with open(fname, "wb") as fd:
        for chunk in r.iter_content(chunk_size=128):
            fd.write(chunk)
    return fname


def str_to_identifier(s, lower=True):
    result = re.sub("\W|^(?=\d)", "_", s)
    if lower:
        result = result.lower()
    return result


# https://gitlab.com/gitlab-org/gitlab/-/blob/master/lib/gitlab/utils.rb#L92
def gitlab_slugify(s):
    s = s.lower()
    s = re.sub(r"[^a-z0-9]", "-", s)[:63]
    s = re.sub(r"(^-+|-+$)", "", s)
    return s


def reponame_to_displayname(reponame: str) -> str:
    return re.sub("[-_]+", " ", reponame).title()


def download_plugin_template(
    template_name: str, url: str = TEMPLATE_URL, base_path: str = TEMPLATE_BASE_PATH
):
    # base_path = Path(base_path) / template_name if url is None else Path(url.rsplit("/", 1)[1].split(".")[0]) / template_name

    zip_path = download_file(url)
    base_path = Path(zip_path.split(".")[0]) / template_name

    with zipfile.ZipFile(zip_path, "r") as f:
        result = {name: f.read(name) for name in f.namelist() if base_path in Path(name).parents}

    if len(result) == 0:
        Path(zip_path).unlink()
        raise ValueError(f"Could not find template: {template_name}")

    result = {
        str(PurePosixPath(k).relative_to(PurePosixPath(base_path))): v.decode("utf-8")
        for k, v in result.items()
        if v
    }
    Path(zip_path).unlink()
    return result


def get_templates(url: str = TEMPLATE_URL) -> List[str]:
    zip_path = download_file(url)

    with zipfile.ZipFile(zip_path, "r") as f:
        files_split = [name.split("/") for name in f.namelist()]
        result = [fn[1] for fn in files_split if fn[-1] == "" and len(fn) == 3]
    return result


class TemplateFormatter:
    def __init__(
        self,
        template_dict: Dict[str, str],
        replace_dict: Dict[str, str],
        tgt_path: Union[str, Path],
        verbose: bool = False,
    ):
        self.template_dict = template_dict
        self.tgt_path = Path(tgt_path)
        self.replace_dict = replace_dict
        self.verbose = verbose

    def format_content(self, content):
        return Template(content).safe_substitute(self.replace_dict)

    def format_path(self, path):
        new_path = Template(path).safe_substitute(self.replace_dict)
        return self.tgt_path / new_path

    def format_file(self, filename, content):
        new_path = self.format_path(filename)
        new_content = self.format_content(content)
        new_path.parent.mkdir(exist_ok=True, parents=True)
        if self.verbose:
            logger.info(f"Formatting {filename} -> {new_path}")
        with open(new_path, "w", encoding="utf-8") as f:
            f.write(new_content)

    def get_files(self):
        return [self.format_path(filename) for filename in self.template_dict.keys()]

    def format(self):
        for filename, content in self.template_dict.items():
            self.format_file(filename, content)

    def print_filetree(self):
        previous_prefix = None
        res = "Created the following files"

        for path in sorted(
            [x.relative_to(self.tgt_path) for x in self.get_files()],
            key=lambda item: 100 * str(item).count("/"),
        ):
            n_slashes = str(path).count("/")
            new_prefix = path.parent
            if previous_prefix != new_prefix and str(new_prefix) != ".":
                res = f"{res}\n├── {new_prefix}"
            if n_slashes == 0:
                res = f"{res}\n├── {path}"
            elif n_slashes == 1:
                res = f"{res}\n│   ├── {path.name}"

            previous_prefix = new_prefix

        print(res.strip() + "\n")


def get_template_replace_dict(
    repo_url=None,
    user=None,
    plugin_name=None,
    package_name=None,
    description=None,
    install_requires=None,
    template_name=None,
):
    if repo_url is None:
        repo_url = get_remote_url()

    try:
        repo_owner, repo_name = infer_git_info(repo_url)
    except ValueError:
        url_inf = None
        logger.error(
            "Could not infer git information from current directory, no initialized repository found."
        )

    if repo_url is None:
        repo_url = url_inf

    if user is None:
        if repo_owner in GITLAB_GROUPS:
            user = None
        else:
            user = repo_owner

    if plugin_name is None:
        if repo_name is None:
            plugin_name = None
        else:
            plugin_name = reponame_to_displayname(repo_name)

    if package_name is None:
        if repo_name is None:
            package_name = None
        else:
            package_name = str_to_identifier(repo_name)
    if install_requires is None:
        install_requires = ""
    else:
        install_requires = "\n    ".join(
            [
                x.strip()
                for x in install_requires.split(",")
                if x.strip() != "" and x.strip() not in ["pymemri", "pytest"]
            ]
        )

    if template_name == "classifier_plugin":
        assert package_name is not None
        assert user is not None
        repo_name_gitlab = gitlab_slugify(repo_name)

        # hacky, dont change!
        model_imports_ = (
            f"""
from transformers import pipeline, AutoTokenizer, AutoModelForSequenceClassification
from """
            + "pymemri.data.loader import load_huggingface_model_for_project"
        )
        model_init = f"""
        model = load_huggingface_model_for_project(project_path="{user}/{repo_name_gitlab}", client=client)
        tokenizer = AutoTokenizer.from_pretrained("distilroberta-base", model_max_length=512)
        self.pipeline = pipeline("sentiment-analysis", model=model, tokenizer=tokenizer, return_all_scores=True, truncation=True)
        """
        model_predict = """
        return self.pipeline(x)
        """
    else:
        model_imports_ = ""
        model_init = "raise NotImplementedError()"
        model_predict = "raise NotImplementedError()"

    return {
        "user": user,
        "package_name": package_name,
        "plugin_name": plugin_name,
        "repo_name": repo_name,
        "repo_url": repo_url,
        "description": str(description),
        "install_requires": install_requires,
        "model_imports": model_imports_,
        "model_init": model_init,
        "model_predict": model_predict,
    }


def _plugin_from_template(
    list_templates=False,
    user=None,
    repo_url=None,
    plugin_name=None,
    template_name="basic",
    package_name=None,
    description=None,
    target_dir=".",
    verbose=True,
    install_requires="",
    template_url=None,
):
    if list_templates:
        print("Available templates:")
        for template in get_templates():
            print(template)
        return

    template = download_plugin_template(
        template_name, **({"url": template_url} if template_url else {})
    )

    tgt_path = Path(target_dir)
    replace_dict = get_template_replace_dict(
        repo_url=repo_url,
        user=user,
        plugin_name=plugin_name,
        package_name=package_name,
        description=description,
        install_requires=install_requires,
        template_name=template_name,
    )
    logger.debug(replace_dict)

    formatter = TemplateFormatter(template, replace_dict, tgt_path)
    formatter.format()
    if verbose:
        formatter.print_filetree()

    logger.info(f"Created `{replace_dict['plugin_name']}` using the {template_name} template.")


@call_parse
def plugin_from_template(
    list_templates: Param("List available plugin templates", store_true) = False,
    user: Param("Your Gitlab username", str) = None,
    repo_url: Param("The url of your empty Gitlab plugin repository", str) = None,
    plugin_name: Param("Display name of your plugin", str) = None,
    template_name: Param(
        "Name of the template, use `list_templates` to see all available options"
    ) = "basic",
    package_name: Param("Name of your plugin python package", str) = None,
    description: Param("Description of your plugin", str) = None,
    target_dir: Param("Directory to output the formatted template", str) = ".",
    verbose: Param("Should print out dir", bool) = True,
    install_requires: Param(
        "Extra packages to install, provided as comma separated, e.g. pymemri,requests", str
    ) = "",
):
    """
    CLI that downloads and formats a plugin template according to the arguments, and local git repository.

    Args:
        list_templates (Param, optional): If True, only list available templates. Defaults to False.
        user (Param, optional): Your GitLab username. Defaults to None.
        repo_url (Param, optional): The url of your gitlab plugin repository. Defaults to None.
        plugin_name (Param, optional): The name of your plugin. Defaults to None.
        template_name (Param, optional): The name of the template used. To list all options, see `list_templates`.
            Defaults to "basic".
        package_name (Param, optional): The name of the python package of your plugin. Inferred if left blank. Defaults to None.
        description (Param, optional): An optional plugin description. Defaults to None.
        target_dir (Param, optional): Directory where the plugin template is generated. Defaults to ".".
    """
    _plugin_from_template(
        list_templates,
        user,
        repo_url,
        plugin_name,
        template_name,
        package_name,
        description,
        target_dir,
        verbose,
        install_requires,
    )



import string
from collections import UserString
from typing import Any, Dict, Optional


class GQLQuery(UserString):
    def __init__(self, query_string: str, variables: Optional[Dict[str, Any]] = None) -> None:
        super().__init__(query_string)
        if variables:
            self.format(variables)

    def format(self, variables: Optional[Dict[str, Any]] = None, **kwargs) -> "GQLQuery":
        self.data = string.Template(self.data).substitute(variables, **kwargs)
        return self



import os
from time import sleep

import requests
from loguru import logger

from pymemri.data.schema import PluginRun
from pymemri.plugin.constants import POD_PLUGIN_DNS_ENV
from pymemri.pod.client import PodClient


def run_qr_flow(client: PodClient, plugin_run: PluginRun):
    host = "0.0.0.0"
    port = plugin_run.webserverPort or 8080
    # give time to start
    logger.info("Checking if webservice is ready...")
    ready = False
    while not ready:
        try:
            ready = requests.get(f"http://{host}:{port}/qr").status_code in [200, 201]
        except Exception as e:
            logger.info(f"The QR server is not ready yet: '{e}'")

        sleep(0.2)

    user_host = os.environ.get(POD_PLUGIN_DNS_ENV, f"http://0.0.0.0:{port}")
    full_user_auth_url = f"{user_host}/qr"

    logger.info(f"GO TO {full_user_auth_url} and scan the code")



from enum import Enum

from loguru import logger

from ..data.schema import Item


class Priority(Enum):
    newest = "newest"
    local = "local"
    remote = "remote"
    error = "error"


class DB:
    def __init__(self):
        self.nodes = dict()

    def add(self, node):
        id = node.id
        if id in self.nodes:
            logger.error(
                f"Error trying to add node, but node with with id: {id} is already in database"
            )
        self.nodes[id] = node

    def get(self, id):
        res = self.nodes.get(id, None)
        return res

    def contains(self, node):
        return node.id in self.nodes

    def merge(self, node, priority):
        if self.contains(node):
            node = self._merge_item(self.get(node.id), node, priority)
        else:
            self.add(node)
        return node

    def _merge_item(self, local_item: Item, remote_item: Item, priority: Priority) -> Item:
        """
        Merge the properties and edges of `remote_item` into `local_item`, according to `priority`

        Possible priorities:
        "newest": In case of a conflict, use the local property
                  if it was modified after remote_item.date_server_modified
        "local": In case of a conflict, use the local property value
        "remote": In case of a conflict, use the remote property value
        "error": throw a `ValueError` on conflict
        """
        if not isinstance(remote_item, type(local_item)):
            raise ValueError(
                f"Attempted to merge two items with different schema: {type(remote_item)}, {type(local_item)}"
            )

        for prop in local_item.properties:
            self._merge_property(local_item, remote_item, prop, priority)

        self._merge_edges(local_item, remote_item)
        return local_item

    def _merge_property(self, local_item, remote_item, prop, priority):
        local_val = getattr(local_item, prop)
        remote_val = getattr(remote_item, prop)
        orig_val = local_item._original_properties.get(prop, None)

        if prop == "dateServerModified":
            setattr(local_item, prop, remote_val)
            return

        # Property is not updated locally since last sync, always use remote
        if prop not in local_item._original_properties:
            setattr(local_item, prop, remote_val)

        elif priority == Priority.newest:
            # Note: Pod does not have a DSM per property, so we compare against the Item DSM.
            dateLocalModified = local_item._date_local_modified.get(prop, None)
            if remote_val != orig_val and remote_item.dateServerModified > dateLocalModified:
                setattr(local_item, prop, remote_val)

        elif priority == Priority.remote:
            if orig_val != remote_val:
                setattr(local_item, prop, remote_val)

        elif priority == Priority.local:
            return

        elif priority == Priority.error:
            raise ValueError(f"Sync conflict on `{prop}` property for {local_item}")

        else:
            raise ValueError(f"Unknown sync priority: {priority}")

    def _merge_edges(self, local_item, remote_item):
        for edge in local_item.edges:
            local_edges = local_item.__edges__[edge]
            remote_edges = remote_item.__edges__[edge]
            local_edges.extend([edge for edge in remote_edges if edge not in local_edges])
            for edge in local_edges:
                edge.source = local_item



import os
import re
import sys
import urllib
from getpass import getpass
from pathlib import Path

import requests
from fastprogress.fastprogress import progress_bar
from git import Repo
from loguru import logger

from pymemri.pod.client import PodClient

from .template.formatter import _plugin_from_template, gitlab_slugify

MEMRI_PATH = Path.home() / ".memri"
MEMRI_GITLAB_BASE_URL = "https://gitlab.memri.io"
ACCESS_TOKEN_PATH = Path.home() / ".memri/access_token/access_token.txt"
GITLAB_API_BASE_URL = "https://gitlab.memri.io/api/v4"
DEFAULT_PACKAGE_VERSION = "0.0.1"

TIME_FORMAT_GITLAB = "%Y-%m-%dT%H:%M:%S.%fZ"
PROJET_ID_PATTERN = '(?<=<span class="gl-button-text">Project ID: )[0-9]+(?=</span>)'


class GitlabAPI:
    def __init__(self, client: PodClient = None, request_auth_if_needed: bool = False):
        self.client = client
        self.auth_headers = dict()
        self.auth_params = dict()
        self.auth_initialized = False
        self.request_auth_if_needed = request_auth_if_needed
        self.get_registry_params_headers()

    def init_auth_params(self):
        access_token = self.client.get_oauth2_access_token("gitlab")
        if access_token is None:
            raise RuntimeError(
                "No gitlab access token found in the pod. "
                "Please authenticate with gitlab in the frontend."
            )
        self.auth_params = {"access_token": access_token}

    def get_registry_params_headers(self):
        job_token = os.environ.get("CI_JOB_TOKEN", None)
        if job_token is not None:
            self.auth_initialized = True
            self.auth_headers = {"JOB-TOKEN": job_token}
        elif self.client is not None:
            self.init_auth_params()
            self.auth_initialized = True
        else:
            ACCESS_TOKEN_PATH.parent.mkdir(parents=True, exist_ok=True)
            if ACCESS_TOKEN_PATH.is_file():
                self.auth_initialized = True
                with open(ACCESS_TOKEN_PATH, "r") as f:
                    self.auth_headers = {"PRIVATE-TOKEN": f.read()}
            else:
                if self.request_auth_if_needed:
                    print(
                        """
                    The first time you are uploading a model you need to create an access_token
                    at https://gitlab.memri.io/-/profile/personal_access_tokens?name=Model+Access+token&scopes=api
                    Click at the blue button with 'Create personal access token'"
                    """
                    )
                    access_token = getpass(
                        "Then copy your personal access token from 'Your new personal access token', and paste here: "
                    )
                    with open(ACCESS_TOKEN_PATH, "w") as f:
                        f.write(access_token)
                    self.auth_headers = {"PRIVATE-TOKEN": access_token}
                    self.auth_initialized = True

        # test if we are authenticated
        res = self.get(
            f"{GITLAB_API_BASE_URL}/projects", headers=self.auth_headers, params=self.auth_params
        )
        if res.status_code not in [200, 201]:
            logger.error(res.content)
            raise RuntimeError(res.content)

    def write_file_to_package_registry(
        self,
        project_id,
        file_path,
        package_name,
        version=DEFAULT_PACKAGE_VERSION,
        trigger_pipeline=False,
    ):
        file_path = Path(file_path)
        file_name = file_path.name

        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{package_name}/{version}/{file_name}"
        logger.info(f"uploading {file_path}")
        it = upload_in_chunks(file_path)
        res = self.put(
            url=url,
            data=IterableToFileAdapter(it),
            headers=self.auth_headers,
            params=self.auth_params,
        )

        if res.status_code not in [200, 201]:
            logger.error(f"Failed to upload {file_path}: {res.content}")
        else:
            logger.info(f"Succesfully uploaded {file_path}")
            if trigger_pipeline:
                self.trigger_pipeline(project_id)

    def _authenticated_request(self, request_function, *args, **kwargs):
        try:
            return request_function(*args, **kwargs)
        except Exception as e:
            logger.error(e)
            if self.auth_params:
                logger.debug("Maybe access_token is expired, trying again with new auth params")
                self.init_auth_params()
                return request_function(*args, **kwargs)
            else:
                raise e

    def get(self, *args, **kwargs):
        return self._authenticated_request(requests.get, *args, **kwargs)

    def post(self, *args, **kwargs):
        return self._authenticated_request(requests.post, *args, **kwargs)

    def put(self, *args, **kwargs):
        return self._authenticated_request(requests.put, *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._authenticated_request(requests.delete, *args, **kwargs)

    def trigger_pipeline(self, project_id):
        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/pipeline?ref=main"
        res = self.post(url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201]:
            logger.error("Failed to trigger pipeline")

    def project_id_from_name(self, project_name, username=None):
        iden = gitlab_slugify(project_name)
        username = username or self.get_current_username()
        uri_encoded_name = urllib.parse.quote_plus(f"{username}/{iden}")
        res = self.get(
            f"{GITLAB_API_BASE_URL}/projects/{uri_encoded_name}",
            headers=self.auth_headers,
            params=self.auth_params,
        )

        if res.status_code not in [200, 201]:
            logger.error(res.content)
            raise RuntimeError(f"Failed to get project id for {project_name}")
        project_id = res.json().get("id", None)
        if project_id:
            return project_id
        else:
            raise ValueError(
                f"No plugin found with name {project_name}, make sure to enter the name as specified in the url of the repo"
            )

    def get_project_id_from_project_path_unsafe(self, project_path):
        try:
            res = self.get(f"{MEMRI_GITLAB_BASE_URL}/{project_path}")
            html = str(res.content)
            match = re.search(PROJET_ID_PATTERN, html)
            return match.group()
        except Exception:
            raise ValueError(f"Could not find project with name {project_path}")

    def download_package_file(
        self,
        filename,
        project_path,
        package_name,
        out_dir=None,
        package_version=DEFAULT_PACKAGE_VERSION,
        download_if_exists=False,
    ):

        project_name = str(project_path).split("/")[-1]
        out_dir = out_dir if out_dir is not None else MEMRI_PATH / "projects" / project_name
        out_dir.mkdir(parents=True, exist_ok=True)

        project_id = self.get_project_id_from_project_path_unsafe(project_path)

        file_path = out_dir / filename

        if file_path.exists() and not download_if_exists:
            logger.warning(
                f"{file_path} already exists, and `download_if_exists`==False, using cached version"
            )
            return file_path

        logger.info(f"downloading {filename} from project {project_path}, package {package_name}")

        res = self.get(
            url=f"{GITLAB_API_BASE_URL}/projects/{project_id}/packages/generic/{package_name}/{package_version}/{filename}"
        )
        res.raise_for_status()
        with open(out_dir / filename, "wb") as f:
            logger.info(f"writing {filename} to {out_dir}")
            f.write(res.content)
        return file_path

    def create_repo(self, repo_name, client=None):
        url = f"{GITLAB_API_BASE_URL}/projects/"
        payload = {"name": repo_name}
        res = self.post(url=url, json=payload, headers=self.auth_headers, params=self.auth_params)

        if res.status_code not in [200, 201]:
            raise ValueError(f"failed to create repo:\n {res.text}")
        logger.info(f"created project {repo_name}")

    def get_current_username(self, client=None):
        url = f"{GITLAB_API_BASE_URL}/user/"
        res = self.get(url=url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201]:
            raise ValueError(f"Could not find current user {res.content}")
        else:
            username = res.json()["username"]
            return username

    # NEVER EXPORT THIS
    def delete_project(self, path_or_id, client=None):
        url_escape_id = urllib.parse.quote(str(path_or_id), safe="")
        url = f"{GITLAB_API_BASE_URL}/projects/{url_escape_id}"
        res = self.delete(url=url, headers=self.auth_headers, params=self.auth_params)
        if res.status_code not in [200, 201, 202]:
            raise ValueError(f"failed to delete repo:\n {res.text}")
        logger.info(f"deleted project {path_or_id}")

    def commit_file(self, project_name, path_in2out, branch="main", client=None):
        project_id = self.project_id_from_name(project_name)
        #     file_out_path_escaped = urllib.parse.quote(file_out_path, safe='')

        actions = []

        for file_in_path, file_out_path in path_in2out.items():
            content = Path(file_in_path).read_text()
            action_payload = {
                "action": "create",
                "file_path": file_out_path,
                "content": content,
            }
            actions.append(action_payload)

        url = f"{GITLAB_API_BASE_URL}/projects/{project_id}/repository/commits"
        payload = {
            "branch": branch,
            "commit_message": "automated commit",
            "content": content,
            "actions": actions,
        }

        res = self.post(url=url, json=payload, headers=self.auth_headers, params=self.auth_params)
        files_in = list(path_in2out.keys())
        if res.status_code not in [200, 201, 202]:
            raise ValueError(f"failed to make commit with files {files_in}:\n {res.text}")
        logger.info(f"committed files {files_in}")

    def write_files_to_git(self, repo, target_dir, **kwargs):
        path_in2out = dict()
        for p in target_dir.rglob("*"):
            if p.is_file():
                path_in_repo = p.relative_to(target_dir)
                path_in2out[str(p)] = str(path_in_repo)
        self.commit_file(str(repo), path_in2out, **kwargs)

    def create_new_project(self, project_name, user=None, template_url=None):
        tmp_dir = Path("/tmp/test") / project_name
        rm_tree(tmp_dir)
        repo_url = (
            f"{MEMRI_GITLAB_BASE_URL}/{user}/{project_name}"
            if user is not None
            else f"{MEMRI_GITLAB_BASE_URL}/plugins/{project_name}"
        )

        _plugin_from_template(
            template_name="classifier_plugin",
            description="A transformer based sentiment analyis plugin",
            install_requires="transformers,sentencepiece,protobuf,torch==1.10.0",
            target_dir=str(tmp_dir),
            repo_url=repo_url,
            verbose=False,
            user=user,
            template_url=template_url,
        )
        self.write_files_to_git(project_name, tmp_dir)


def find_git_repo():
    path = "."
    for i in range(10):
        try:
            repo = Repo(f"{path + ('.' * i)}/")
        except:
            pass
        else:
            break
    if i == 9:
        raise ValueError(f"could not fine git repo in {os.path.abspath('')}")

    repo_name = repo.remotes.origin.url.split(".git")[0].split("/")[-1]
    return repo_name


class upload_in_chunks(object):
    def __init__(self, filename, chunksize=1 << 14):
        self.filename = filename
        self.chunksize = chunksize
        self.totalsize = os.path.getsize(filename)
        self.readsofar = 0

    def __iter__(self):
        n = 100
        pb = progress_bar(range(n))
        pb_iter = iter(pb)
        i = 1
        delta = 1 / n
        next(pb_iter, None)

        with open(self.filename, "rb") as file:
            while True:
                data = file.read(self.chunksize)
                if not data:
                    sys.stderr.write("\n")
                    break
                self.readsofar += len(data)
                percent = self.readsofar * 1e2 / self.totalsize
                while (percent / 100) > i * delta:
                    next(pb_iter, None)
                    i += 1
                yield data
        pb.update_bar(n)

    def __len__(self):
        return self.totalsize


class IterableToFileAdapter(object):
    def __init__(self, iterable):
        self.iterator = iter(iterable)
        self.length = len(iterable)

    def read(self, size=-1):  # TBD: add buffer for `len(data) > size` case
        return next(self.iterator, b"")

    def __len__(self):
        return self.length


def rm_tree(pth):
    pth = Path(pth)
    for child in pth.glob("*"):
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)

    try:
        pth.rmdir()
    except FileNotFoundError as e:
        pass



from pathlib import Path
from typing import List

import pymemri

from ..data.schema import CVUStoredDefinition

CVU_BASE_PATH = Path(pymemri.__file__).parent / "cvu" / "definitions"


def get_default_cvu(name: str, base_path: Path = CVU_BASE_PATH) -> CVUStoredDefinition:
    """
    Get a CVU by name.
    """
    path = Path(base_path) / name
    cvu_str = Path(path).read_text()
    return CVUStoredDefinition(definition=cvu_str, name=name, externalId=name)


def list_default_cvus(base_path: Path = CVU_BASE_PATH) -> List[str]:
    """
    List all CVUs available in pymemri.
    """
    path = Path(base_path)
    cvus = path.glob("*.cvu")
    cvu_names = [path.name for path in cvus]
    return cvu_names



import http.server
import socketserver
import urllib
from urllib.parse import parse_qs, urlsplit

from pymemri.data.schema import OauthFlow
from pymemri.pod.client import PodClient


def get_request_handler(
    client: PodClient, oauth_token_secret: str
) -> http.server.BaseHTTPRequestHandler:
    """
    This is a factory function that returns a request handler class.

    The returned class will have a reference to the client and oauth_token_secret
    variables that are passed to this function.

    This is needed because the request handler class is instantiated by the
    TCPServer class, and we need to pass the client and oauth_token_secret
    variables to the request handler class.
    """

    class MyHttpRequestHandler(http.server.SimpleHTTPRequestHandler):
        def do_GET(self):
            params = urlsplit(self.path)
            if params.path == "/oauth":
                args = parse_qs(params.query)
                oauth_verifier = args.get("oauth_verifier")[0]
                oauth_token = args.get("oauth_token")[0]
                response = client.get_oauth1_access_token(
                    oauth_token=oauth_token,
                    oauth_verifier=oauth_verifier,
                    oauth_token_secret=oauth_token_secret,
                )
                access_token = response["oauth_token"]
                access_token_secret = response["oauth_token_secret"]
                item = OauthFlow(
                    service="twitter",
                    accessToken=access_token,
                    accessTokenSecret=access_token_secret,
                )
                client.create(item)

                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(bytes("Authenticated, succesfully created oauth item", "utf-8"))

    return MyHttpRequestHandler


def run_twitter_oauth_flow(
    *,
    client: PodClient,
    callback_url: str,
    host: str = "localhost",
    port: int = 3667,
) -> None:
    callback_url = callback_url or f"http://{host}:{port}/oauth?state=twitter"
    response = client.get_oauth1_request_token("twitter", callback_url)

    oauth_token_secret = response["oauth_token_secret"]
    queryParameters = {"oauth_token": response["oauth_token"]}
    encoded = urllib.parse.urlencode(queryParameters)
    print(f"*** \n\nGo to https://api.twitter.com/oauth/authorize?{encoded} \n\n***\n\n")

    socketserver.TCPServer.allow_reuse_address = True
    my_server = socketserver.TCPServer(
        (host, port), get_request_handler(client, oauth_token_secret)
    )

    client.add_to_schema(OauthFlow)
    my_server.handle_request()



import inspect
from typing import Any, Dict

from pydantic import validator
from pydantic.fields import Field

from ._central_schema import *  # noqa
from .dataset import Dataset  # noqa
from .photo import Photo  # noqa
from .utils import resolve_forward_refs


def get_schema() -> Dict[str, type]:
    return {k: v for k, v in globals().items() if inspect.isclass(v) and issubclass(v, Item)}


def get_schema_cls(cls_name: str, extra: Optional[Dict[str, type]] = None) -> type:
    """Returns the schema model for `cls_name` if it is in central schema or extra.

    If both central schema and extra contain a class with the same `__name__`, extra
    has the priority.

    Args:
        cls_name (str): Name of the schema model to return the class for
        extra (Dict[str, type], optional): _description_. Defaults to None.

    Raises:
        TypeError: `cls_name` is not a known schema model.

    Returns:
        ItemType: subclass of ItemBase
    """
    if extra and cls_name in extra:
        return extra[cls_name]

    pymemri_schema = get_schema()
    if cls_name in pymemri_schema:
        return pymemri_schema[cls_name]
    else:
        raise TypeError(f"{cls_name} is not a known schema model")


class PluginRun(PluginRun):
    @validator("id", always=True, allow_reuse=True)
    def validate_has_run_id(
        cls, value: Any, values: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> Any:
        if value is None:
            value = cls.create_id()
        return value

    @validator("targetItemId", always=True, allow_reuse=True)
    def validate_id_is_targetItemId(
        cls, value: Any, values: Dict[str, Any], **kwargs: Dict[str, Any]
    ) -> Any:
        uid = values.get("id")
        if value is None:
            value = uid

        if value != uid:
            raise ValueError("targetItemId should match id")

        return value


# Required for typechecking on forward refs in schema
# Leave at bottom of schema.py
resolve_forward_refs(get_schema())



from inspect import getfullargspec

ENDPOINT_METADATA = "__endpoint_metadata__"


def register_endpoint(endpoint_name: str, method: str):
    """Registers `fn` under endpoint `endpoint_name` into the webserver as public API.
    The `method` parameter specifies HTTP operation: GET, POST, DELETE, etc.
    The `fn` needs to pass the validation, for example cannot contain *args, **kwargs,
    all arguments needs to have type annotation."""

    def _add_endpoint_metadata_attr(fn):
        args = getfullargspec(fn)

        # *args makes fastapi crash upon calling the handler
        if args.varargs:
            raise RuntimeError(f"passed function cannot have *args")

        # Allowing **kwargs:
        # - leaks python detail to the API
        # - is treated as string in query parameter: "{"key": value, ...}", instead of {"key": value}
        if args.varkw:
            raise RuntimeError(f"passed function cannot have **kwargs")

        # Requiring typing enforces api clarity
        missing_annotations = []
        for arg in args.args:
            if arg != "self" and arg not in args.annotations:
                missing_annotations.append(arg)

        if missing_annotations:
            raise RuntimeError(f"arguments: {missing_annotations} does not have type annotation")

        metadata = (endpoint_name, method)

        # Prevent nested decoration
        if hasattr(fn, ENDPOINT_METADATA):
            raise RuntimeError(
                f"Trying to register function {fn.__name__} to {metadata} but is already used in: {fn.__endpoint_metadata__}"
            )

        fn.__endpoint_metadata__ = metadata
        return fn

    return _add_endpoint_metadata_attr



import inspect
import json
import os
from pathlib import Path
from typing import List

from fastcore.script import Param, call_parse
from loguru import logger

from ..plugin.pluginbase import get_plugin_cls
from ..pod.client import PodClient

ALLOWED_TYPES = [int, str, float, bool]


def get_params(cls):
    params = inspect.signature(cls.__init__).parameters
    return {k: v for k, v in list(params.items()) if k not in {"self", "args", "kwargs"}}


def identifier_to_displayname(identifier: str) -> str:
    return identifier.replace("_", " ").title()


def get_param_config(name, dtype, is_optional, default):
    return {
        "name": name,
        "display": identifier_to_displayname(name),
        "data_type": dtype,
        "type": "textbox",
        "default": default,
        "optional": is_optional,
    }


def create_config(plugin_cls: type) -> List[dict]:
    """
    Returns a declarative plugin configuration, inferred from the `__init__` method signature of `plugin_cls`.
    This function is used internally by the `create_plugin_config` CLI. For general use, use the CLI instead.

    Arguments that start with `_`, untyped arguments or arguments that are not in `ALLOWED_TYPES` are skipped.

    Args:
        plugin_cls (type): A plugin class, inherited from PluginBase.

    Returns:
        List[dict]: A declarative configuration definition as list of dictionaries.
    """
    config = list()
    for param_name, param in get_params(plugin_cls).items():
        if param_name.startswith("_"):
            continue
        if param.annotation == inspect._empty:
            logger.info(f"Skipping unannotated parameter `{param_name}`")
            continue
        if param.annotation not in ALLOWED_TYPES:
            logger.info(f"Skipping parameter with unknown type: `{param_name}: {param.annotation}`")
            continue
        is_optional = param.default != inspect._empty
        dtype = PodClient.TYPE_TO_SCHEMA[param.annotation]
        default = param.default if is_optional else None
        param_config = get_param_config(param_name, dtype, is_optional, default)
        config.append(param_config)
    return config


@call_parse
def create_plugin_config(
    metadata: Param("metadata.json of the plugin", str) = "./metadata.json",
    tgt_file: Param("Filename of config file", str) = "config.json",
    schema_file: Param("Filename of exported plugin schema", str) = "schema.json",
):
    """
    Creates a plugin configuration definition from the arguments of your plugin class.

    Configuration arguments are inferred from the arguments of your plugin `__init__` method.
    Arguments that start with `_`, untyped arguments or arguments that are not in `ALLOWED_TYPES` are skipped.
    All generated fields are "textbox" by default, in the future our front-end will support more
    types of fields.
    Args:
        metadata (Param, optional): Location of the "metadata.json" file,
            Defaults to "./metadata.json"
        tgt_file (Param, optional): File the config definition is saved to.
            Defaults to "config.json".
    """
    if metadata is None:
        if os.path.exists("./metadata.json"):
            metadata = "./metadata.json"
        else:
            print("Define a metadata file with --metadata <filename>")
            return
    with open(metadata, "r") as f:
        metadata = json.load(f)

    plugin_module = metadata["pluginModule"]
    plugin_name = metadata["pluginName"]

    try:
        plugin_cls = get_plugin_cls(plugin_module, plugin_name)
    except Exception as e:
        logger.error(e)
        return
    config = create_config(plugin_cls)

    with open(tgt_file, "w") as f:
        json.dump(config, f, indent=2)
    print(f"Config saved to {Path(tgt_file)}")

    plugin_schema = plugin_cls.get_schema()
    with open(schema_file, "w") as f:
        json.dump(plugin_schema, f, indent=2)



import abc
import threading

from loguru import logger

from pymemri.webserver.models.trigger import TriggerReq

from .pluginbase import PluginBase


class TriggerPluginBase(PluginBase):
    """Base class for plugins that are expected to expose HTTP interface used by the POD.
    Current use case is to notify the plugin about arrival of new data."""

    def __init__(self, pluginRun=None, client=None, **kwargs):
        """The pluginRun argument keeps information about port used to setup the web server"""
        super().__init__(pluginRun=pluginRun, client=client, **kwargs)

        # Pass a closure to the fastapi route
        self._webserver.app.add_api_route("/v1/item/trigger", self.do_trigger, methods=["POST"])

        self.daemon = True

    def do_trigger(self, req: TriggerReq):
        """Handle trigger request for given item. Item must be present already in the POD.
        Operation is offloaded to a dedicated thread, the POD is notified about the status
        asynchronously."""

        def thread_fn(req: TriggerReq):
            try:
                self.trigger(req)
                self.client.send_trigger_status(req.item_ids, req.trigger_id, "OK")

            except Exception as e:
                msg = f"Error while handling the trigger {req.trigger_id}, reason {e}"
                logger.error(msg)
                self.client.send_trigger_status(req.item_ids, req.trigger_id, msg)

        threading.Thread(target=thread_fn, args=(req,)).start()

        return "ok"

    @abc.abstractmethod
    def trigger(self, req: TriggerReq):
        """
        Handler of new data arrival, argument holds required details
        @throws Exception in any error encountered
        """
        raise NotImplementedError()



import sys
from collections import UserList
from typing import Any, Dict, ForwardRef, Tuple, Union

if sys.version_info >= (3, 9, 0):
    from typing import get_args, get_origin

    def evaluate_type(type_: Any, localns) -> Any:
        return type_._evaluate(globalns=None, localns=localns, recursive_guard=set())

else:

    def get_args(type_annotation: Any) -> Any:
        return getattr(type_annotation, "__args__", tuple())

    def get_origin(type_annotation: Any) -> Any:
        return getattr(type_annotation, "__origin__", None)

    def evaluate_type(type_: Any, localns) -> Any:
        return type_._evaluate(globalns=None, localns=localns)


def type_to_str(_type: type) -> str:
    """Returns string representation of `_type`, ForwardRef is supported."""
    if isinstance(_type, type):
        return _type.__name__
    elif isinstance(_type, ForwardRef):
        return _type.__forward_arg__


def type_or_union_to_tuple(_type: type) -> Tuple[type]:
    """converts Union[TypeA, TypeB] to (TypeA, TypeB)

    If _type is not a union, return (_type,)
    """
    if get_origin(_type) == Union:
        return tuple(get_args(_type))
    return (_type,)


def update_union_forward_refs(field, schema):
    # Standard pydantic resolver does not handle Union[ForwardRef(A), ForwardRef(B)] -> Union[A, B]
    # TODO log as issue with Pydantic
    if get_origin(field.type_) == Union:
        new_args = tuple(
            evaluate_type(arg, localns=schema)
            for arg in get_args(field.type_)
            if isinstance(arg, ForwardRef)
        )
        field.type_.__args__ = new_args


def resolve_forward_refs(schema: Dict[str, type]):
    """
    Resolve forward refs for all classes in schema.values()
    """
    for schema_cls in list(schema.values()):
        schema_cls.update_forward_refs(**schema)
        for field in schema_cls.__edge_fields__.values():
            update_union_forward_refs(field, schema)



import json
import os
import urllib
from collections import deque
from hashlib import sha256
from typing import Any, Deque, Dict, Generator, List, Optional, Union

import requests
from loguru import logger

from .graphql_utils import GQLQuery

DEFAULT_POD_ADDRESS = os.environ.get("POD_ADDRESS") or "http://localhost:3030"
POD_VERSION = "v4"


class PodError(Exception):
    def __init__(self, status=None, message=None, **kwargs) -> None:
        super().__init__(status, message, **kwargs)
        self.status = status
        self.message = message

    def __str__(self) -> str:
        return " ".join([str(a) for a in self.args if a])


class PodAPI:
    def __init__(
        self,
        database_key: str,
        owner_key: str,
        url: str = DEFAULT_POD_ADDRESS,
        version: str = POD_VERSION,
        auth_json: dict = None,
        verbose: bool = True,
    ) -> None:
        self.verbose = verbose
        self.database_key = database_key
        self.owner_key = owner_key
        self.version = version
        self._url = url
        self.base_url = f"{url}/{version}/{self.owner_key}"
        self.auth_json = self._create_auth(auth_json)
        self.session = requests.Session()

    def _create_auth(self, auth_json: dict = None) -> dict:
        if auth_json is not None:
            return {"type": "PluginAuth", **auth_json}
        else:
            return {"type": "ClientAuth", "databaseKey": self.database_key}

    def create_account(self):
        response = self.session.post(
            f"{self._url}/{self.version}/account",
            json={"ownerKey": self.owner_key, "databaseKey": self.database_key},
        )
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)

    def test_connection(self) -> bool:
        try:
            res = self.session.get(self._url)
            if self.verbose:
                logger.info("Succesfully connected to pod")
            return True
        except requests.exceptions.RequestException as e:
            logger.error("Could no connect to backend")
            return False

    @property
    def pod_version(self) -> dict:
        response = self.session.get(f"{self._url}/version")
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response.json()

    def post(self, endpoint: str, payload: Any) -> Any:
        body = {"auth": self.auth_json, "payload": payload}
        response = self.session.post(f"{self.base_url}/{endpoint}", json=body)
        if response.status_code != 200:
            raise PodError(response.status_code, response.text)
        return response

    def get_item(self, uid: str) -> dict:
        return self.post("get_item", uid).json()

    def create_item(self, item: dict) -> str:
        return self.post("create_item", item).json()

    def update_item(self, item: dict) -> list:
        return self.post("update_item", item).json()

    def get_edges(
        self, uid: str, direction: str = "Outgoing", expand_items: bool = True
    ) -> List[dict]:
        payload = {"item": uid, "direction": direction, "expandItems": expand_items}
        return self.post("get_edges", payload).json()

    def create_edge(self, edge: dict) -> str:
        return self.post("create_edge", edge).json()

    def delete_item(self, uid) -> list:
        return self.post("delete_item", uid).json()

    def search(self, query: dict) -> List[dict]:
        return self.post("search", query).json()

    def search_paginate(self, query: dict, limit: int = 32, even_page_size=True) -> Generator:
        """
        The Pod returns uneven page sizes when paginating, which can be an issue for some applications.
        `search_paginate` wraps the pagination, and always returns pages of size `limit` by storing overflow items
        in a queue.
        """
        paginator = self._paginate(query, limit)

        if not even_page_size:
            yield from paginator

        remaining: Deque[Dict[str, Any]] = deque()
        while True:
            if len(remaining) >= limit:
                yield [remaining.popleft() for _ in range(limit)]
            if len(remaining) < limit:
                try:
                    remaining.extend(next(paginator))
                except StopIteration:
                    break

        while len(remaining):
            yield [remaining.popleft() for _ in range(min(limit, len(remaining)))]

    def _paginate(self, query: dict, limit: int = 32) -> Generator:
        if (
            "_limit" in query
            or "dateServerModified" in query
            or "dateServerModified>=" in query
            or "dateServerModified<" in query
        ):
            raise ValueError("Cannot paginate query that contains a date or limit.")
        if "_sortOrder" in query:
            raise NotImplementedError("Only 'Asc' order is supported.")

        query = {**query, "_limit": limit}
        response = self.search(query)
        if not len(response):
            return
            yield

        next_date = 0
        while True:
            query["dateServerModified>="] = next_date
            response = self.search(query)
            if not len(response):
                break

            next_date = response[-1]["dateServerModified"] + 1
            yield response

    def bulk(
        self,
        create_items: List[dict] = None,
        update_items: List[dict] = None,
        create_edges: List[dict] = None,
        delete_items: List[str] = None,
        search: List[dict] = None,
    ) -> Dict[str, Any]:

        payload = {
            "createItems": create_items,
            "updateItems": update_items,
            "createEdges": create_edges,
            "deleteItems": delete_items,
            "search": search,
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self.post("bulk", payload).json()

    def graphql(
        self, query: Union[str, GQLQuery], variables: Optional[Dict[str, Any]] = None
    ) -> List[dict]:
        if isinstance(query, str):
            query = GQLQuery(query)
        query.format(variables)
        return self.post("graphql", query.data).json()

    def oauth(self, service: str, callback_url: str):
        payload = {"service": service, "callbackUrl": callback_url}
        return self.post("oauth", payload).json()

    def upload_file(self, file: bytes) -> Any:
        if self.auth_json.get("type") == "PluginAuth":
            # alternative file upload for plugins, with different authentication
            return self.upload_file_b(file)

        sha = sha256(file).hexdigest()
        result = self.session.post(
            f"{self.base_url}/upload_file/{self.database_key}/{sha}", data=file
        )
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def upload_file_b(self, file: bytes) -> Any:
        sha = sha256(file).hexdigest()
        auth = urllib.parse.quote(json.dumps(self.auth_json))
        result = self.session.post(f"{self.base_url}/upload_file_b/{auth}/{sha}", data=file)
        if result.status_code != 200:
            raise PodError(result.status_code, result.text)

        return result

    def get_file(self, sha: str) -> bytes:
        return self.post("get_file", {"sha256": sha}).content

    def send_email(self, to: str, subject: str = "", body: str = "") -> Any:
        payload = {"to": to, "subject": subject, "body": body}
        return self.post("send_email", payload)

    def send_trigger_status(self, item_ids: List[str], trigger_id: str, status: str) -> Any:
        payload = {"item_ids": item_ids, "trigger_id": trigger_id, "status": status}
        return self.post("trigger/status", payload)

    def oauth2get_access_token(
        self,
        platform: str,
    ) -> Dict:
        return self.post(
            "oauth2/access_token",
            {"platform": platform},
        ).json()

    def oauth2authorize(
        self, *, platform: str, code: str, redirect_uri: str, pkce_verifier: str
    ) -> Dict[str, str]:
        return self.post(
            "oauth2/authorize",
            {
                "platform": platform,
                "authCode": code,
                "redirectUri": redirect_uri,
                "pkceVerifier": pkce_verifier,
            },
        ).json()

    def oauth2get_authorization_url(
        self, platform: str, scopes: str, redirect_uri: str
    ) -> Dict[str, str]:
        return self.post(
            "oauth2/auth_url",
            {
                "platform": platform,
                "scopes": scopes,
                "redirectUri": redirect_uri,
            },
        ).json()

    def oauth1_request_token(self, platform: str, callback_url: str) -> Any:
        return self.post(
            "oauth1_request_token", {"service": platform, "callbackUrl": callback_url}
        ).json()

    def oauth1_access_token(self, *, oauth_token, oauth_token_secret, oauth_verifier) -> Any:
        return self.post(
            "oauth1_access_token",
            {
                "oauthVerifier": oauth_verifier,
                "oauthToken": oauth_token,
                "oauthTokenSecret": oauth_token_secret,
            },
        ).json()

    def plugin_status(self, plugins: List[str]) -> Any:
        return self.post("plugin/status", {"plugins": plugins}).json()

    def plugin_api(self, plugin_id: str) -> Any:
        return self.post("plugin/api", {"id": plugin_id}).json()

    def plugin_api_call(
        self,
        plugin_id: str,
        method: str,
        endpoint: str,
        query: Dict[str, Any] = None,
        jsonBody: Dict[str, Any] = None,
    ) -> Any:
        payload = {
            "id": plugin_id,
            "method": method,
            "endpoint": endpoint,
            "query": query,
            "jsonBody": jsonBody,
        }

        return self.post("plugin/api/call", payload).json()
